# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

"""
Model that wraps all segmentation models of mmdetection
"""
import logging
from math import degrees
from typing import Any, Dict, List, Optional, Tuple, Union, cast

import torch.nn
from mlcvzoo_base.api.data.annotation_class_mapper import AnnotationClassMapper
from mlcvzoo_base.api.data.bounding_box import BoundingBox
from mlcvzoo_base.api.data.box import Box
from mlcvzoo_base.api.data.class_identifier import ClassIdentifier
from mlcvzoo_base.api.data.types import ImageType
from mlcvzoo_base.api.interfaces import NetBased, Trainable
from mlcvzoo_base.api.model import ObjectDetectionModel
from mlcvzoo_base.api.structs import Runtime
from mlcvzoo_base.configuration.structs import ObjectDetectionBBoxFormats
from mlcvzoo_base.configuration.utils import (
    create_configuration as create_basis_configuration,
)
from mlcvzoo_mmdetection.configuration import (
    MMDetectionConfig,
    MMDetectionInferenceConfig,
    MMDetectionMMDeployConfig,
)
from mlcvzoo_mmdetection.mlcvzoo_mmdet_dataset import MLCVZooMMDetDataset
from mlcvzoo_mmdetection.model import MMDetectionModel
from mmdet.apis import inference_detector
from mmdet.datasets.transforms.formatting import PackDetInputs
from mmdet.datasets.transforms.loading import LoadAnnotations
from mmrotate.registry import DATASETS, TRANSFORMS
from mmrotate.structures.bbox import RotatedBoxes
from mmrotate.structures.bbox.quadri_boxes import QuadriBoxes
from torch import Tensor

try:
    from mlcvzoo_mmdetection.mlcvzoo_mmdeploy.inferencer import MMDeployInferencer
except ModuleNotFoundError:
    MMDeployInferencer = None  # type: ignore # pylint: disable=invalid-name
try:
    from mlcvzoo_mmdetection.mlcvzoo_mmdeploy.converter import MMDeployConverter
except ModuleNotFoundError:
    MMDeployConverter = None  # type: ignore # pylint: disable=invalid-name


logger = logging.getLogger(__name__)


class MMRotateModel(
    MMDetectionModel[MMDetectionInferenceConfig],
    ObjectDetectionModel[MMDetectionConfig, Union[str, ImageType]],
    NetBased[torch.nn.Module, MMDetectionInferenceConfig],
    Trainable,
):
    """
    Class for wrapping mmrotate rotated object detection models
    """

    def __init__(
        self,
        from_yaml: Optional[str] = None,
        configuration: Optional[MMDetectionConfig] = None,
        string_replacement_map: Optional[Dict[str, str]] = None,
        init_for_inference: bool = False,
        is_multi_gpu_instance: bool = False,
        runtime: str = Runtime.DEFAULT,
    ) -> None:
        self.mmdeploy_inferencer: Optional[MMDeployInferencer] = None

        MMDetectionModel.__init__(
            self,
            configuration=self.create_configuration(
                from_yaml, configuration, string_replacement_map
            ),
            string_replacement_map=string_replacement_map,
            init_for_inference=init_for_inference,
            is_multi_gpu_instance=is_multi_gpu_instance,
            runtime=runtime,
        )
        ObjectDetectionModel.__init__(
            self,
            configuration=self.configuration,
            mapper=AnnotationClassMapper(
                class_mapping=self.configuration.class_mapping,
                reduction_mapping=self.configuration.inference_config.reduction_class_mapping,
            ),
            init_for_inference=init_for_inference,
            runtime=runtime,
            is_rotation_model=True,
        )
        NetBased.__init__(self, net=self.net)
        Trainable.__init__(self)

    @staticmethod
    def create_configuration(
        from_yaml: Optional[str] = None,
        configuration: Optional[MMDetectionConfig] = None,
        string_replacement_map: Optional[Dict[str, str]] = None,
    ) -> MMDetectionConfig:
        return cast(
            MMDetectionConfig,
            create_basis_configuration(
                configuration_class=MMDetectionConfig,
                from_yaml=from_yaml,
                input_configuration=configuration,
                string_replacement_map=string_replacement_map,
            ),
        )

    @property
    def num_classes(self) -> int:
        return self.mapper.num_classes

    def get_classes_id_dict(self) -> Dict[int, str]:
        return self.mapper.annotation_class_id_to_model_class_name_map

    def _init_inference_model(self) -> None:
        from mmdet.apis import init_detector

        if self.runtime == Runtime.DEFAULT:
            if self.net is None:
                self.net = init_detector(
                    self.cfg,
                    device=self.configuration.inference_config.device_string,
                )
                if self.configuration.inference_config.checkpoint_path != "":
                    self.restore(
                        checkpoint_path=self.configuration.inference_config.checkpoint_path
                    )
        elif self.runtime in (
            Runtime.ONNXRUNTIME,
            Runtime.ONNXRUNTIME_FLOAT16,
            Runtime.TENSORRT,
        ):
            # Return value of _get__mmdeploy_config will not be None
            mmdeploy_config = cast(
                MMDetectionMMDeployConfig, self._get_mmdeploy_config()
            )

            if MMDeployInferencer is None:
                raise RuntimeError(
                    "Extra 'mmdeploy' must be installed to run a model which is deployed with "
                    "MMDeploy."
                )

            self.mmdeploy_inferencer = MMDeployInferencer(
                model_config=self.cfg,
                mmdeploy_config=mmdeploy_config,
            )
        else:
            raise ValueError(
                f"Initialization for inference is not supported for runtime '{self.runtime}'."
            )

    @staticmethod
    def _register_dataset() -> None:
        """
        Register the custom dataset of the MLCVZoo in the registry of mmrotate

        Returns:
            None
        """

        TRANSFORMS.register_module(
            LoadAnnotations.__name__, module=LoadAnnotations, force=True
        )
        TRANSFORMS.register_module(
            PackDetInputs.__name__, module=PackDetInputs, force=True
        )
        DATASETS.register_module(
            MLCVZooMMDetDataset.__name__, module=MLCVZooMMDetDataset, force=True
        )

    def __decode_mmdet_prediction(
        self,
        prediction: Any,
    ) -> List[BoundingBox]:
        """
        Decode output of an object detection model from mmdetection

        Args:
            prediction: The result that the model has predicted

        Returns:
            The decoded prediction as list of bounding boxes in MLCVZoo format
        """

        bounding_boxes: List[BoundingBox] = []

        if isinstance(prediction.pred_instances.bboxes, Tensor):
            if prediction.pred_instances.bboxes.size(-1) == 5:
                mmdet_bboxes = RotatedBoxes(prediction.pred_instances.bboxes)
            elif prediction.pred_instances.bboxes.size(-1) == 8:
                mmdet_bboxes = QuadriBoxes(prediction.pred_instances.bboxes)
                mmdet_bboxes.convert_to("rbox")
            else:
                raise TypeError(
                    "Require the shape of `bboxes` to be (n, 5) "
                    "or (n, 8), but get `bboxes` with shape being "
                    f"{prediction.pred_instances.bboxes.shape}."
                )
        elif isinstance(prediction.pred_instances.bboxes, RotatedBoxes):
            mmdet_bboxes = prediction.pred_instances.bboxes
        else:
            raise ValueError(
                "prediction.pred_instances.bboxes needs to be a Tensor for decoding"
            )

        for mmdet_bbox, score_tensor, label_tensor in zip(
            mmdet_bboxes,
            prediction.pred_instances.scores,
            prediction.pred_instances.labels,
        ):
            score = float(score_tensor)
            if score < self.configuration.inference_config.score_threshold:
                continue

            class_id = int(label_tensor)

            class_identifiers = (
                self.mapper.map_model_class_id_to_output_class_identifier(
                    class_id=class_id
                )
            )

            mmdet_bbox_list = mmdet_bbox.cpu().tensor.reshape(-1).tolist()

            model_class_identifier = ClassIdentifier(
                class_id=class_id,
                class_name=self.mapper.map_annotation_class_id_to_model_class_name(
                    class_id=class_id
                ),
            )

            for class_identifier in class_identifiers:
                bounding_boxes.append(
                    BoundingBox(
                        class_identifier=class_identifier,
                        model_class_identifier=model_class_identifier,
                        score=float(score),
                        box=Box.init_format_based(
                            box_list=mmdet_bbox_list[0:4],
                            box_format=ObjectDetectionBBoxFormats.CXCYWH,
                            angle=degrees(float(mmdet_bbox_list[4])),
                        ),
                        difficult=False,
                        occluded=False,
                        content="",
                        background=False,
                    )
                )

        return bounding_boxes

    def predict(
        self, data_item: Union[str, ImageType]
    ) -> Tuple[Union[str, ImageType], List[BoundingBox]]:
        if self.runtime == Runtime.DEFAULT:
            if self.net is None:
                raise ValueError(
                    "The 'net' attribute is not initialized, "
                    "make sure to instantiate with init_for_inference=True"
                )

            prediction = inference_detector(model=self.net, imgs=data_item)
            return data_item, self.__decode_mmdet_prediction(prediction=prediction)

        if self.runtime in (
            Runtime.ONNXRUNTIME,
            Runtime.ONNXRUNTIME_FLOAT16,
            Runtime.TENSORRT,
        ):
            if self.mmdeploy_inferencer is None:
                raise ValueError(
                    "The 'inferencer' attribute is not initialized, make sure to instantiate with "
                    "init_for_inference=True"
                )

            if MMDeployInferencer is None:
                raise RuntimeError(
                    "Extra 'mmdeploy' must be installed to run a model which is deployed with "
                    "MMDeploy."
                )

            return data_item, self.__decode_mmdet_prediction(
                prediction=self.mmdeploy_inferencer(data_item)[0]
            )

        else:
            raise ValueError(
                f"Prediction is not supported for runtime '{self.runtime}'."
            )

    def predict_many(
        self, data_items: List[Union[str, ImageType]]
    ) -> List[Tuple[Union[str, ImageType], List[BoundingBox]]]:

        prediction_list: List[Tuple[Union[str, ImageType], List[BoundingBox]]] = []

        if self.runtime == Runtime.DEFAULT:
            if self.net is None:
                raise ValueError(
                    "The 'net' attribute is not initialized, "
                    "make sure to instantiate with init_for_inference=True"
                )

            predictions = inference_detector(model=self.net, imgs=data_items)

        elif self.runtime in (
            Runtime.ONNXRUNTIME,
            Runtime.ONNXRUNTIME_FLOAT16,
            Runtime.TENSORRT,
        ):
            if self.mmdeploy_inferencer is None:
                raise ValueError(
                    "The 'inferencer' attribute is not initialized, make sure to instantiate with "
                    "init_for_inference=True"
                )
            if MMDeployInferencer is None:
                raise RuntimeError(
                    "Extra 'mmdeploy' must be installed to run a model which is deployed with "
                    "MMDeploy."
                )

            predictions = [
                self.mmdeploy_inferencer(inputs=data_item)[0]
                for data_item in data_items
            ]

        else:
            raise ValueError(
                f"Multi-prediction is not supported for runtime '{self.runtime}'."
            )

        for data_item, prediction in zip(data_items, predictions):
            bounding_boxes = self.__decode_mmdet_prediction(prediction=prediction)
            prediction_list.append(
                (
                    data_item,
                    bounding_boxes,
                )
            )
        return prediction_list


if __name__ == "__main__":
    MMDetectionModel.run(MMRotateModel)
